/* tslint:disable:no-use-before-declare */
"use strict";
var __extends = (this && this.__extends) || function (d, b) {
    for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p];
    function __() { this.constructor = d; }
    d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
};
require('reflect-metadata');
/**
 * Represents optional values. Instances of `Optional`
 *  are either an instance of $some or the object $none.
 *
 *  The most idiomatic way to use an $optional instance is to treat it
 *  as a collection or monad and use `map`,`flatMap`, `filter`, or
 *  `foreach`:
 */
var Optional = (function () {
    function Optional() {
    }
    /**
     * Returns the option's value if the option is nonempty, otherwise
     * return the result of evaluating `default`.
     *
     * @param a the default expression.
     */
    Optional.prototype.getOrElse = function (a) {
        return this.isEmpty ? a : this.get();
    };
    /**
     * Returns a $some containing the result of applying $f to this $option's value if this $option is nonempty.
     * Otherwise return $none.
     *
     * @note This is similar to `flatMap` except here, $f does not need to wrap its result in an $option.
     *
     * @param f the function to apply
     * @see flatMap
     * @see foreach
     */
    Optional.prototype.map = function (f) {
        return this.isEmpty ? exports.None : Some(f(this.get()));
    };
    /**
     * Returns the result of applying $f to this $option's
     * value if the $option is nonempty.  Otherwise, evaluates
     * expression `ifEmpty`.
     *
     * @note This is equivalent to `$option map f getOrElse ifEmpty`.
     *
     * @param ifEmpty the expression to evaluate if empty.
     * @param f       the function to apply if nonempty.
     */
    Optional.prototype.fold = function (ifEmpty, f) {
        return this.isEmpty ? ifEmpty : f(this.get());
    };
    /**
     * Returns the result of applying $f to this $option's value if this $option is nonempty.
     * Returns $none if this $option is empty.
     * Slightly different from `map` in that $f is expected to return an $option (which could be $none).
     *
     * @param f the function to apply
     * @see map
     * @see foreach
     */
    Optional.prototype.flatMap = function (f) {
        return this.isEmpty ? exports.None : f(this.get());
    };
    /**
     * Returns this $option if it is nonempty '''and''' applying the predicate $F to
     * this $option's value returns true. Otherwise, return $none.
     *
     * @param p the predicate used for testing.
     */
    Optional.prototype.filter = function (f) {
        return this.isEmpty || f(this.get()) ? this : exports.None;
    };
    /**
     * Tests whether the option contains a given value as an element.
     *
     * @param elem the element to test.
     * @return `true` if the option has an element that is equal (as determined by `==`) to `elem`, `false` otherwise.
     */
    Optional.prototype.contains = function (b) {
        return this.nonEmpty && this.get() === b;
    };
    /**
     * Returns true if this option is nonempty '''and''' the predicate $f returns true when applied to this $option's value.
     * Otherwise, returns false.
     *
     * @param p the predicate to test
     */
    Optional.prototype.exists = function (f) {
        return this.nonEmpty && f(this.get());
    };
    /**
     * Returns true if this option is empty '''or''' the predicate $f returns true when applied to this $option's value.
     *
     * @param p the predicate to test
     */
    Optional.prototype.forall = function (f) {
        return this.isEmpty || f(this.get());
    };
    /**
     * Apply the given procedure $f to the option's value, if it is nonempty. Otherwise, do nothing.
     *
     * @param f the procedure to apply.
     * @see map
     * @see flatMap
     */
    Optional.prototype.foreach = function (f) {
        if (this.nonEmpty) {
            f(this.get());
        }
    };
    /**
     * Returns this $option if it is nonempty, otherwise return the result of evaluating `alternative`.
     *
     * @param alternative the alternative expression.
     */
    Optional.prototype.orElse = function (ob) {
        return this.isEmpty ? ob : this;
    };
    // add methods
    Optional.prototype.apply1 = function (ob, f) {
        return this.flatMap(function (a) { return ob.map(function (b) { return f(a, b); }); });
    };
    Optional.prototype.apply2 = function (ob, oc, f) {
        return this.flatMap(function (a) { return ob.flatMap(function (b) { return oc.map(function (c) { return f(a, b, c); }); }); });
    };
    Optional.prototype.apply3 = function (ob, oc, od, f) {
        return this.flatMap(function (a) { return ob.flatMap(function (b) { return oc.flatMap(function (c) { return od.map(function (d) { return f(a, b, c, d); }); }); }); });
    };
    Optional.prototype.apply4 = function (ob, oc, od, oe, f) {
        return this.flatMap(function (a) { return ob.flatMap(function (b) { return oc.flatMap(function (c) { return od.flatMap(function (d) { return oe.map(function (e) { return f(a, b, c, d, e); }); }); }); }); });
    };
    Optional.prototype.apply5 = function (ob, oc, od, oe, of, f) {
        return this.flatMap(function (a) { return ob.flatMap(function (b) { return oc.flatMap(function (c) { return od.flatMap(function (d) { return oe.flatMap(function (e) { return of.map(function (ff) { return f(a, b, c, d, e, ff); }); }); }); }); }); });
    };
    Optional.prototype.chain = function (ob) {
        return new OptionalBuilder1(this, ob);
    };
    Optional.apply = function (a) {
        return (a !== undefined && a !== null) ? new SomeImpl(a) : exports.None;
    };
    return Optional;
}());
exports.Optional = Optional;
/**
 * Class `Some[A]` represents existing values of type `A`.
 */
function Some(a) {
    return new SomeImpl(a);
}
exports.Some = Some;
//------------------------------------
//
// Implementation section.
//
//------------------------------------
var SomeImpl = (function (_super) {
    __extends(SomeImpl, _super);
    function SomeImpl(value) {
        _super.call(this);
        this.value = value;
        this.isEmpty = false;
        this.nonEmpty = true;
    }
    SomeImpl.prototype.get = function () {
        return this.value;
    };
    SomeImpl.prototype.toString = function () {
        return 'Some(' + this.value + ')';
    };
    return SomeImpl;
}(Optional));
var NoneImpl = (function (_super) {
    __extends(NoneImpl, _super);
    function NoneImpl() {
        _super.apply(this, arguments);
        this.isEmpty = true;
        this.nonEmpty = false;
    }
    NoneImpl.prototype.get = function () {
        throw new TypeError('None can not #get');
    };
    NoneImpl.prototype.toString = function () {
        return 'None';
    };
    return NoneImpl;
}(Optional));
/**
 * This object represents non-existent values.
 */
exports.None = new NoneImpl();
// Builders
var OptionalBuilder1 = (function () {
    function OptionalBuilder1(oa, ob) {
        this.oa = oa;
        this.ob = ob;
    }
    OptionalBuilder1.prototype.run = function (f) {
        return this.oa.apply1(this.ob, f);
    };
    OptionalBuilder1.prototype.chain = function (oc) {
        return new OptionalBuilder2(this.oa, this.ob, oc);
    };
    return OptionalBuilder1;
}());
exports.OptionalBuilder1 = OptionalBuilder1;
var OptionalBuilder2 = (function () {
    function OptionalBuilder2(oa, ob, oc) {
        this.oa = oa;
        this.ob = ob;
        this.oc = oc;
    }
    OptionalBuilder2.prototype.run = function (f) {
        return this.oa.apply2(this.ob, this.oc, f);
    };
    OptionalBuilder2.prototype.chain = function (od) {
        return new OptionalBuilder3(this.oa, this.ob, this.oc, od);
    };
    return OptionalBuilder2;
}());
exports.OptionalBuilder2 = OptionalBuilder2;
var OptionalBuilder3 = (function () {
    function OptionalBuilder3(oa, ob, oc, od) {
        this.oa = oa;
        this.ob = ob;
        this.oc = oc;
        this.od = od;
    }
    OptionalBuilder3.prototype.run = function (f) {
        return this.oa.apply3(this.ob, this.oc, this.od, f);
    };
    OptionalBuilder3.prototype.chain = function (oe) {
        return new OptionalBuilder4(this.oa, this.ob, this.oc, this.od, oe);
    };
    return OptionalBuilder3;
}());
exports.OptionalBuilder3 = OptionalBuilder3;
var OptionalBuilder4 = (function () {
    function OptionalBuilder4(oa, ob, oc, od, oe) {
        this.oa = oa;
        this.ob = ob;
        this.oc = oc;
        this.od = od;
        this.oe = oe;
    }
    OptionalBuilder4.prototype.run = function (f) {
        return this.oa.apply4(this.ob, this.oc, this.od, this.oe, f);
    };
    OptionalBuilder4.prototype.chain = function (of) {
        return new OptionalBuilder5(this.oa, this.ob, this.oc, this.od, this.oe, of);
    };
    return OptionalBuilder4;
}());
exports.OptionalBuilder4 = OptionalBuilder4;
var OptionalBuilder5 = (function () {
    function OptionalBuilder5(oa, ob, oc, od, oe, of) {
        this.oa = oa;
        this.ob = ob;
        this.oc = oc;
        this.od = od;
        this.oe = oe;
        this.of = of;
    }
    OptionalBuilder5.prototype.run = function (f) {
        return this.oa.apply5(this.ob, this.oc, this.od, this.oe, this.of, f);
    };
    return OptionalBuilder5;
}());
exports.OptionalBuilder5 = OptionalBuilder5;
